@@ -14,6 +14,7 @@
 #include <rpm/rpmdb.h>
 #include <rpm/rpmfi.h>
 #include <rpm/rpmts.h>
+#include <rpm/rpmsq.h>
 #include <rpm/rpmlog.h>
 #include <rpm/rpmfileutil.h>	/* rpmCleanPath */
 
@@ -146,8 +147,8 @@ int showQueryPackage(QVA_t qva, rpmts ts, Header h)
 	if ((qva->qva_flags & QUERY_FOR_LICENSE) && !(fflags & RPMFILE_LICENSE))
 	    continue;
 
-	/* If not querying %ghost, skip ghost files. */
-	if ((qva->qva_fflags & RPMFILE_GHOST) && (fflags & RPMFILE_GHOST))
+	/* Skip on attributes (eg from --noghost) */
+	if (fflags & qva->qva_fflags)
 	    continue;
 
 	if (qva->qva_flags & QUERY_FOR_STATE) {
@@ -261,7 +262,6 @@ void rpmDisplayQueryTags(FILE * fp)
 	}
 	fprintf(fp, "\n");
     }
-    rpmtdFreeData(names);
     rpmtdFree(names);
 }
 
@@ -273,7 +273,7 @@ static int rpmgiShowMatches(QVA_t qva, rpmts ts, rpmgi gi)
     while ((h = rpmgiNext(gi)) != NULL) {
 	int rc;
 
-	rpmdbCheckSignals();
+	rpmsqPoll();
 	if ((rc = qva->qva_showPackage(qva, ts, h)) != 0)
 	    ec = rc;
 	headerFree(h);
@@ -291,7 +291,7 @@ static int rpmcliShowMatches(QVA_t qva, rpmts ts, rpmdbMatchIterator mi)
 
     while ((h = rpmdbNextIterator(mi)) != NULL) {
 	int rc;
-	rpmdbCheckSignals();
+	rpmsqPoll();
 	if ((rc = qva->qva_showPackage(qva, ts, h)) != 0)
 	    ec = rc;
     }
@@ -304,7 +304,7 @@ static rpmdbMatchIterator initQueryIterator(QVA_t qva, rpmts ts, const char * ar
     int i;
     rpmdbMatchIterator mi = NULL;
 
-    (void) rpmdbCheckSignals();
+    (void) rpmsqPoll();
 
     if (qva->qva_showPackage == NULL)
 	goto exit;
@@ -482,7 +482,9 @@ static rpmdbMatchIterator initQueryIterator(QVA_t qva, rpmts ts, const char * ar
 	}
 	mi = rpmdbFreeIterator(mi);
 	if (! matches) {
-	    rpmlog(RPMLOG_NOTICE, _("package %s is not installed\n"), arg);
+	    size_t l = strlen(arg);
+	    if (!(l > 4 && !strcmp(arg + l - 4, ".rpm")))
+		rpmlog(RPMLOG_NOTICE, _("package %s is not installed\n"), arg);
 	} else {
 	    mi = rpmtsInitIterator(ts, RPMDBI_LABEL, arg, 0);
 	}
@@ -547,6 +549,7 @@ int rpmcliArgIter(rpmts ts, QVA_t qva, ARGV_const_t argv)
 	break;
     }
     case RPMQV_SPECRPMS:
+    case RPMQV_SPECBUILTRPMS:
     case RPMQV_SPECSRPM:
 	for (ARGV_const_t arg = argv; arg && *arg; arg++) {
 	    ec += ((qva->qva_specQuery != NULL)
@@ -557,6 +560,14 @@ int rpmcliArgIter(rpmts ts, QVA_t qva, ARGV_const_t argv)
 	for (ARGV_const_t arg = argv; arg && *arg; arg++) {
 	    rpmdbMatchIterator mi = initQueryIterator(qva, ts, *arg);
 	    ec += rpmcliShowMatches(qva, ts, mi);
+	    if (mi == NULL && qva->qva_source == RPMQV_PACKAGE) {
+		size_t l = strlen(*arg);
+		if (l > 4 && !strcmp(*arg + l - 4, ".rpm")) {
+		    rpmgi gi = rpmgiNew(ts, giFlags, argv);
+		    ec += rpmgiShowMatches(qva, ts, gi);
+		    rpmgiFree(gi);
+		}
+	    }
 	    rpmdbFreeIterator(mi);
 	}
 	break;
