@@ -130,7 +130,7 @@ void rpmtriggersPrepPostUnTransFileTrigs(rpmts ts, rpmte te)
 
     if (rpmdbGetIteratorCount(mi)) {
 	/* Filter triggers and save only trans postun triggers into ts */
-	while((trigH = rpmdbNextIterator(mi)) != NULL) {
+	while ((trigH = rpmdbNextIterator(mi)) != NULL) {
 	    rpmdsTriggers = rpmdsNew(trigH, RPMTAG_TRANSFILETRIGGERNAME, 0);
 	    while ((rpmdsTrigger = rpmdsFilterTi(rpmdsTriggers, tix))) {
 		if ((rpmdsNext(rpmdsTrigger) >= 0) &&
@@ -171,18 +171,19 @@ int runPostUnTransFileTrigs(rpmts ts)
 	trigH = rpmdbGetHeaderAt(rpmtsGetRdb(ts),
 				trigs->triggerInfo[i].hdrNum);
 
-	/* Maybe package whith this trigger is already uninstalled */
+	/* Maybe package with this trigger is already uninstalled */
 	if (trigH == NULL)
 	    continue;
 
 	/* Prepare and run script */
-	script = rpmScriptFromTriggerTag(trigH, RPMSENSE_TRIGGERPOSTUN,
+	script = rpmScriptFromTriggerTag(trigH,
+		triggertag(RPMSENSE_TRIGGERPOSTUN),
 		RPMSCRIPT_TRANSFILETRIGGER, trigs->triggerInfo[i].tix);
 
 	headerGet(trigH, RPMTAG_INSTPREFIXES, &installPrefixes,
 		HEADERGET_ALLOC|HEADERGET_ARGV);
 
-	nerrors += runScript(ts, NULL, installPrefixes.data, script, 0, 0);
+	nerrors += runScript(ts, NULL, trigH, installPrefixes.data, script, 0, 0);
 	rpmtdFreeData(&installPrefixes);
 	rpmScriptFree(script);
 	headerFree(trigH);
@@ -423,7 +424,7 @@ static int runHandleTriggersInPkg(rpmts ts, rpmte te, Header h,
 	    inputFunc = (char *(*)(void *)) matchFilesNext;
 	    rpmScriptSetNextFileFunc(script, inputFunc, mfi);
 
-	    nerrors += runScript(ts, te, installPrefixes.data,
+	    nerrors += runScript(ts, te, h, installPrefixes.data,
 				script, 0, 0);
 	    rpmtdFreeData(&installPrefixes);
 	    rpmScriptFree(script);
