@@ -1,5 +1,4 @@
 #include "system.h"
-const char *__progname;
 
 #include <rpm/rpmbuild.h>
 #include <rpm/argv.h>
@@ -8,8 +7,6 @@ const char *__progname;
 
 #include "debug.h"
 
-char *progname;
-
 static int print_provides;
 
 static int print_requires;
@@ -75,10 +72,7 @@ main(int argc, char *argv[])
     int ec = 1;
     char buf[BUFSIZ];
 
-    if ((progname = strrchr(argv[0], '/')) != NULL)
-	progname++;
-    else
-	progname = argv[0];
+    xsetprogname(argv[0]); /* Portability call -- see system.h */
 
     optCon = rpmcliInit(argc, argv, optionsTable);
     if (optCon == NULL)
@@ -107,7 +101,7 @@ main(int argc, char *argv[])
 	goto exit;
 
     if (_rpmfc_debug)
-	rpmfcPrint(buf, fc, NULL);
+	rpmfcPrint(NULL, fc, NULL);
 
     if (print_provides)
 	rpmdsPrint(NULL, rpmfcProvides(fc), stdout);
